
#TODO -- turn these into proper JSON strings, not python Dicts!!

SERVICE_SUMMARY_MOCK = {
    "documentId": "683",
    "services": [
        {
            "c": "101",
            "dvbtriplet": "2.2045.6301",
            "schedule": True,
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One Lon",
            "xsg": 3
        },
        {
            "c": "0214",
            "dvbtriplet": "2.2095.55128",
            "servicetype": "DSAT",
            "sf": "au",
            "sg": 20,
            "sid": "2862",
            "sk": 2862,
            "t": "UCB Ireland",
            "xsg": 4
        }
    ],
    "version": 3
}

SERVICE_DETAIL_1 = {
    "details": {
        "dvbtriplet": "2.2045.6301",
        "isbroadcasting": True,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    },
    "sid": "2002",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ts"
        }
    ]
}

SERVICE_DETAIL_2 = {
    "details": {
        "dvbtriplet": "2.2095.55128",
        "isbroadcasting": True,
        "upgradeMessage": "A channel for God's Peace."
    },
    "sid": "2862",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3AB2E/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3AB2E/profileAnExample.ts"
        }
    ]
}

SERVICE_MOCK = [SERVICE_SUMMARY_MOCK,SERVICE_DETAIL_1,SERVICE_DETAIL_2]

REMOTE_TCP_MOCK = [  # set up the data to be returned on each successive call of socket.recv()
    b'SKY 000.001\n',
    b'\x01\x01',
    b'\x00\x00\x00\x00',
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
]


WS_STATUS_MOCK = """
{
   "camessage" : {
      "reason" : "no message",
      "state" : "unavailable"
   },
   "drmstatus" : {
      "state" : "available"
   },
   "entitlements" : [
      "ANALYTICS",
      "BIGBASIC",
      "ETHAN_APP_1",
      "HD",
      "PDL",
      "SKY_DRM_CE",
      "SKY_DRM_MR",
      "SKY_IPPV",
      "ULTRA+",
      "SKY+",
      "GATEWAYENABLER",
      "SIDELOAD"
   ],
   "epginfobits" : {
      "epginfobits" : "0xFDE5FFC0",
      "mask" : "0x5FFA003F",
      "state" : "available"
   },
   "gatewayservices" : {
      "state" : "available"
   },
   "hdmi" : {
      "2160p10bitCapable" : false,
      "authenticatedHDCP" : "NONE",
      "reason" : "HDMI output port is disabled",
      "sinkHDCP" : "NONE",
      "sinkHLG" : false,
      "sinkUHD" : false,
      "state" : "unavailable",
      "uhdConfigured" : false
   },
   "network" : {
      "state" : "available"
   },
   "nssplayback" : {
      "state" : "available"
   },
   "pvr" : {
      "state" : "available"
   },
   "schedule" : {
      "lastdate" : "20181005",
      "state" : "available"
   },
   "servicelist" : {
      "state" : "available"
   },
   "smartcard" : {
      "active" : true,
      "bouquet" : "4101",
      "countryCode" : "GBR",
      "currency" : "GBP",
      "cwe" : true,
      "householdid" : "10947783",
      "paired" : true,
      "state" : "available",
      "subbouquet" : "1",
      "transactionlimit" : 65535,
      "viewingCardNumber" : "725 325 260"
   },
   "swupdate" : {
      "reason" : "IDLE",
      "state" : "unavailable"
   },
   "systemupdates" : {
      "entitlements" : 2,
      "install" : 1,
      "servicegenres" : 1,
      "smartcard" : 1
   },
   "updatetask" : {
      "reason" : "no update",
      "state" : "unavailable"
   }
}


"""



"""Import and merge all modules in this package."""
import logging

from .constants import RCMD
from ._version import __version__
from .remote import press_remote
from .epg import EPG
from .status import Status

logging.getLogger(__name__).addHandler(logging.NullHandler())



#!/usr/bin/env python
"""
Cli wrapper for the skyq API.

Example:
    You can invoke the cli client like this:

        ``$ python cli_epg.py pause``

"""

import argparse
import sys
from typing import List

from pyskyq import EPG


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse command line parameters"""
    parser = argparse.ArgumentParser(
        description="Query Channel Description from SkyQ EPG")
    parser.add_argument(
        dest="sid",
        help="channel sid, try 2002 as an example",
        type=str,
        metavar="SID")
    return parser.parse_args(args)


def main(args: List[str]):
    """Main entry point allowing external calls"""
    pargs = parse_args(args)
    epg = EPG('skyq')  # replace with hostname / IP of your Sky box
    epg.load_channel_data() # load channel listing from Box.

    print(epg.get_channel(pargs.sid).desc)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
"""
Example use of the Status API.

"""

import logging
import sys

from pyskyq import Status

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

logging.getLogger().setLevel(logging.DEBUG)  # status changes are logged.

stat = Status('skyq')  # replace with hostname / IP of your Sky box
stat.create_event_listener()  # set up listener thread.

# do other stuff.

# standby property will be updated asynchronously when the box is turned on or off.
if stat.standby:
    print('The SkyQ Box is in Standby Mode')
else:
    print('The SkyQ Box is in Online Mode')

stat.shudown_event_listener()  # shut down listener thread.



""" This module implements the Status class.

It implements an :py:mod:`asyncio`-based event loop in a separate thread. It will then
schedule subscriptions/polling tasks that subscribe to various events in async
co-routines/tasks and add them to the event loop. Each of these will await an event
(e.g. on a websocket or uPNP subscription) and then respond when neccessary.

"""

import asyncio
import functools
import json
import logging
import signal
import socket
from threading import Thread
from typing import Dict

import websockets

from .constants import REST_PORT, REST_STATUS_URL

LOGGER = logging.getLogger(__name__)

# pylint: disable=too-many-instance-attributes
# For now this will do until standby gets refactored into a dict of items.

class Status:
    """ This class implements the Status object, an internal :py:mod:`asyncio`-based
    event loop used to subscribe to various web-socket and/or uPNP based events.

    It also logs all status changes.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        port (int): Port number to use to connect to the REST HTTP server.
            Defaults to the standard port used by SkyQ boxes which is 9006.
        ws_url (str): ``ws://`` based URL to the SkyQ Box websocket.
        standby (bool): A Boolean indicator stating whether the box is in
            Standby Mode or not.


    """
    def __init__(self,
                 host: str,
                 *,
                 port: int = REST_PORT,
                 ws_url_path: str = REST_STATUS_URL,
                 ws_timeout: int = 20,
                 ping_timeout: int = 10,
                 ) -> None:
        """Initialise the Status object.

        Args:
            host (str): String with resolvable hostname or IPv4 address to SkyQ box.
            port (int, optional): Port number to use to connect to the Remote REST API.
                Defaults to the standard port used by SkyQ boxes which is ``9006``.
            ws_url_path (str, optional): Path stiring to append to the URL, defaults to
                ``/as/system/status``
            ws_timeout (int, optional): Web socket connection timeout. Defaults is 20 sec.
            ping_timeout (int, optional): Web socket ping timeout. Defaults is 10 sec.

        Returns:
            None

        """

        self.host: str = host
        self.port: int = port
        self.standby: bool = False
        # TODO turn into @property and make non-writeable.
        self.ws_url: str = f'ws://{self.host}:{self.port}{ws_url_path}'
        self._event_thread: Thread  # the thread which runs the asyncio event loop
        self._event_loop: asyncio.AbstractEventLoop   # the event loop which runs in the thread.
        self._shutdown_sentinel: bool = False # used to trigger a clean shutdown.
        self._ws_timeout: int = ws_timeout
        self._ping_timeout: int = ping_timeout
        LOGGER.debug(f"Initialised Status object object with host={host}, port={port}")


    async def _ws_subscribe(self) -> None:
        """ Fetch data from web socket asynchronously.

        This  method fetches data from a web socket asynchronously. It is used to
        fetch subscribe to websockets of the SkyQ box.

        """
        LOGGER.debug(f'Setting up web socket listener on {self.ws_url}.')

        while not self._shutdown_sentinel:  # while not being told to shut down
        # outer loop that will restart every time the connection fails (if sentinel says its okay)
            LOGGER.debug('No shutdown sentinel set, so (re)-starting websocket connection.')
            try:
                async with websockets.connect(self.ws_url) as ws:

                    while not self._shutdown_sentinel:
                        # listener loop
                        LOGGER.debug('Starting websocket listener loop iteration...')
                        try:
                            LOGGER.debug('Waiting for data...')
                            payload = await asyncio.wait_for(ws.recv(), timeout=self._ws_timeout)
                            LOGGER.debug(f'Web-socket data received. size = {len(payload)}')
                        except (asyncio.TimeoutError, websockets.exceptions.ConnectionClosed) as err:  # pylint: disable=line-too-long
                            LOGGER.debug(f'Websocket timed out or was closed. Error = {err}')
                            try:
                                if self._shutdown_sentinel:
                                    LOGGER.debug(f'Shutdown triggered. Shutting down...')
                                    break # inner listening loop.
                                else:
                                    LOGGER.debug(f'Trying ping message...')
                                    pong = await ws.ping()
                                    await asyncio.wait_for(pong, timeout=self._ping_timeout)
                                    LOGGER.debug(f'Ping OK, keeping connection alive.')
                            except:  # pylint: disable=bare-except
                                LOGGER.debug(f'Ping timeout - retrying...')
                                break # inner listener loop
                        # process payload
                        LOGGER.debug(f'Invoking payload handler on received message...')
                        asyncio.create_task(self._handle(json.loads(payload)))

            except (socket.gaierror, ConnectionRefusedError) as sc_err:
                await asyncio.sleep(1)
                LOGGER.debug(f'Could not connect to web socket. Error={sc_err}. Retrying...')
                # sleep a bit, log it, then try again
                continue



    def create_event_listener(self) -> None:
        """Create asyncio event loop thread for status event handling.

        Note:
            This method spawns a separate thread and in that thread it runs
            an :py:mod:`asyncio` based event loop. This event loop manages
            all subscriptions to various endpoints/services on the SkyQ box
            and takes the appropriate update actions when these events fire.

            This means that you will be able to simply refer to properties
            like :attr:`standby` and rely on them to report the _current_
            status from the SkyQ Box.

        """

        def _start_event_loop_thread() -> None:
            """Kick off a new asyncio eventloop in a dedicated thread.

            This method implements the thread. """
            LOGGER.debug(f"Asyncio event loop thread running...")
            self._shutdown_sentinel = False
            asyncio.set_event_loop(self._event_loop)
            self._event_loop.run_forever()

        # create a new loop
        self._event_loop = asyncio.new_event_loop()
        self._event_loop.add_signal_handler(signal.SIGTERM,
                                            functools.partial(asyncio.ensure_future,  # pylint: disable=line-too-long
                                                              self._shutdown_signal_handler(signal.SIGTERM)))  # pylint: disable=line-too-long

        # Assign the loop to another thread
        self._event_thread = Thread(target=_start_event_loop_thread, daemon=True)
        self._event_thread.start()

        asyncio.run_coroutine_threadsafe(
            self._ws_subscribe(),
            self._event_loop
        )


    async def _handle(self,
                      payload: Dict,
                      ) -> None:
        """Update status properties with payload."""
        self.standby = payload['hdmi']['state'] != 'available'
        LOGGER.info(f'standby =  {str(self.standby)}')


    async def _shutdown_signal_handler(self,
                                       sig: int,
                                       ) -> None:
        """Shut down event loop cleanly."""


        # mypy gets confused wit IntEnums
        LOGGER.info(f'Caught {sig.name} signal')  # type: ignore
        self._shutdown_sentinel = True  # trigger websocked to shutdown cleanly.
        tasks = [task for task in asyncio.Task.all_tasks() if task is not
                 asyncio.tasks.Task.current_task()]
        results = await asyncio.gather(*tasks, return_exceptions=True).cancel()  # type: ignore
        LOGGER.info(f'Finished awaiting cancelled tasks, results = {results}')
        self._event_loop.stop()
        self._event_loop.close()

    def shudown_event_listener(self) -> None:
        """Terminate the event listener thread and event loop."""
        self._shutdown_sentinel = True



# pylint: skip-file

import asyncio
import logging
import sys
import time

import pytest
import websockets
from asynctest import CoroutineMock, MagicMock

from pyskyq import Status

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import WS_STATUS_MOCK

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

logging.getLogger().setLevel(logging.DEBUG)

def serve_ws_json_with_detail():
    time.sleep(0.2)
    return WS_STATUS_MOCK


def test_status(mocker):
    a = mocker.patch('websockets.connect', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.recv = \
        CoroutineMock(side_effect=serve_ws_json_with_detail)

    stat = Status('some_host')
    stat.create_event_listener()

    time.sleep(0.5) # allow time for awaiting, etc.
    assert stat.standby is True

    stat.shudown_event_listener()


timeout_test_call_count :int  # global var to could calls.

def server_then_close():
    global timeout_test_call_count
    timeout_test_call_count += 1
    print(timeout_test_call_count)
    time.sleep(0.2)
    if timeout_test_call_count > 5:
        raise websockets.exceptions.ConnectionClosed
    else:
        return WS_STATUS_MOCK


def test_status_timeout(mocker):
    global timeout_test_call_count
    timeout_test_call_count = 0
    a = mocker.patch('websockets.connect', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.recv = \
        CoroutineMock(side_effect=server_then_close)

    stat = Status('timeout_host', ws_timeout=2)
    stat.create_event_listener()

    time.sleep(2)

    assert stat.standby is True

    stat.shudown_event_listener()


def test_status_shutdown_sentinel(mocker):
    global timeout_test_call_count
    timeout_test_call_count = 0
    a = mocker.patch('websockets.connect', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.recv = \
        CoroutineMock(side_effect=server_then_close)

    stat = Status('shutdown_host')
    stat.create_event_listener()
    time.sleep(1)
    stat.shudown_event_listener()
    assert stat._shutdown_sentinel is True
    time.sleep(2)
    assert stat.standby is True
